from unittest import TestCase
from betterlist import BetterList as BL


class TestBetterList(TestCase):
    def test_copy(self):
        _list1 = BL([1, 2, 3, 4])
        self.assertEqual(_list1, _list1.copy())

    def test_sort(self):
        _list1 = BL([2, 3, 1, 0])
        _sorted_list1 = BL([0, 1, 2, 3])
        self.assertEqual(_list1.sort(), _sorted_list1)

        _sorted_list2 = BL([3, 1, 0, 2])
        self.assertEqual(_list1.sort(lambda x: (-2) ** x), _sorted_list2)

    def test_min(self):
        _list1 = BL([2, 3, 1, 4])
        self.assertEqual(_list1.min(1)[0], 1)
        self.assertEqual(_list1.min(2, lambda x: (-2) ** x), [3, 1])

    def test_max(self):
        _list1 = BL([2, 3, 1, 4, 5])
        self.assertEqual(_list1.max(1)[0], 5)
        self.assertEqual(_list1.max(2, lambda x: (-2) ** x), [4, 2])

    def test_value_after(self):
        _list1 = BL([2, 1, 4, 5, 3])
        self.assertEqual(_list1.value_after(1), 4)

    def test_first(self):
        _list1 = BL([2, 1, 4, 5])
        self.assertEqual(_list1.first(3), BL([2, 1, 4]))

    def test_create_list(self):
        _list1 = BL(range(3))
        self.assertEqual(_list1, BL([0, 1, 2]))

    def test_add(self):
        _list1 = BL(range(5))
        self.assertEqual(_list1, BL([0, 1, 2, 3]) + BL([4]))
        self.assertEqual(type(_list1), type(BL([0, 1, 2, 3]) + BL([4])))
        _list1 += BL([5])
        self.assertEqual(_list1, BL([0, 1, 2, 3]) + BL([4, 5]))
        self.assertEqual(type(_list1), type(BL([0, 1, 2, 3]) + BL([4, 5])))

    def test_mul(self):
        _list1 = BL(range(3)) * 3
        _result = BL([0, 1, 2, 0, 1, 2, 0, 1, 2])
        self.assertEqual(_list1, _result)
        self.assertEqual(type(_list1), type(_result))
        _list1 = BL(range(3))
        _list1 *= 3
        _result = BL([0, 1, 2, 0, 1, 2, 0, 1, 2])
        self.assertEqual(_list1, _result)
        self.assertEqual(type(_list1), type(_result))

    def test_getitem(self):
        list1 = BL(range(4))
        self.assertEqual(list1[1], 1)
        self.assertEqual(list1[0:2], BL([0, 1]))
        self.assertEqual(type(list1[0:2]), BL)
        self.assertEqual(list1[::-1], BL([3, 2, 1, 0]))
        self.assertEqual(type(list1[::-1]), BL)



import asyncio
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Callable, TypeVar
import nest_asyncio
from betterlist import BetterList

nest_asyncio.apply()
T = TypeVar("T")


def run_concurrent_tasks(async_func: Callable[[T], Any], args_list: list[T]):
    """
    This function is shortcut for the situation where you want to run function
    over a list concurrently.

    You need to have:

    1. list of inputs
        Ex.
        num_list = [1,2,3,4]

    2. async function that takes single input from the args list
        Ex.
        async def print_num(num:int):
            await asyncio.sleep(1)
            print(num)
            return num

    Usage
        Ex.
        results = run_concurrent_tasks(print_num,num_list)
        print(results)
    """

    async def main():
        tasks = BetterList([asyncio.create_task(async_func(arg)) for arg in args_list])
        results = await asyncio.gather(*tasks)
        return results

    return asyncio.run(main())


@dataclass
class Message:
    ok: bool
    time: datetime
    sender: str
    category: str
    message: Any


class Producer:
    """
    The producer of messages.
    Modify at async run function
    Use
    - self.send_msg to send normal message
    - self.send_err to send error message

    Ex.

    class MyProducer(Producer):
        async def run(self):
            while True:
                await asyncio.sleep(1)
                await self.send_msg("NormalMSG", "This is message!")


    producer_1 = MyProducer(name="1")
    """

    def __init__(self, name=""):
        self.notify_corr: Callable[[Message], Any]
        self.name = name

    async def send_msg(self, category: str, message: Any):
        msg = Message(
            ok=True,
            time=datetime.now(),
            sender=self.name,
            category=category,
            message=message,
        )
        await self.notify_corr(msg)

    async def send_err(self, category: str, message: Any):
        err = Message(
            ok=False,
            time=datetime.now(),
            sender=self.name,
            category=category,
            message=message,
        )
        await self.notify_corr(err)

    async def run(self):
        raise NotImplementedError


class Consumer:
    """
    The consumer of messages.
    Modify at async on_msg function

    Ex.
    class MyConsumer(Consumer):
        async def on_msg(msg):
            if msg.category != "MyCategory":
                return
            elif not msg.ok:
                print(f"Error! {msg}")
                return
            print(f"I got my message! {msg}")
    """

    def __init__(self, name=""):
        self.name = name

    async def on_msg(self, msg: Message):
        raise NotImplementedError


class MessageChannel:
    """
    The communication channel.
    There can only be one message channel at a time in an application.
    There're 4 classes working together.

                             Channel
    Producer#1  ---Message---> | |  ---Message--->  Every Consumer
                               | |
    Producer#2  ---Message---> | |  ---Message--->  Every Consumer
    Producer#1  ---Message---> | |  ---Message--->  Every Consumer
                               | |

    Ex.
    Channel = MessageChannel()
    Channel .register_producer(producer_1)
            .register_producer(producer_2)
            .register_consumer(consumer_1)
            .register_consumer(consumer_2)
            .run()
    """

    def __init__(self):
        self.producers: list[Producer] = []
        self.consumers: list[Consumer] = []

    async def notify_corr(self, msg: Message):
        tasks = BetterList(
            [asyncio.create_task(con.on_msg(msg)) for con in self.consumers]
        )
        return await asyncio.gather(*tasks)

    def register_producer(self, producer: Producer):
        producer.notify_corr = self.notify_corr
        self.producers.append(producer)
        return self

    def register_consumer(self, consumer: Consumer):
        self.consumers.append(consumer)
        return self

    def run(self):
        async def main():
            tasks = BetterList(
                [asyncio.create_task(producer.run()) for producer in self.producers]
            )
            return await asyncio.gather(*tasks)

        return asyncio.run(main())



from functools import reduce
from typing import Any


def compose(*func):
    def compose_function(f, g):
        return lambda *args: g(f(*args))

    return reduce(compose_function, func)


class Result:
    def __init__(self, value: Any) -> None:
        self.__value = value
        self.is_ok: bool

    def unwrap(self):
        if not self.is_ok:
            raise ValueError(self.__value)
        return self.__value

    def unwrap_or_default(self, default: Any):
        if not self.is_ok:
            raise default
        return self.__value

    def expect(self, reason: str):
        """
        Describe the reason you expect the Result should be Ok.
        """
        if not self.is_ok:
            raise ValueError(reason)
        return self.__value


class Ok(Result):
    def __init__(self, value: Any) -> None:
        super().__init__(value)
        self.is_ok = True


class Err(Result):
    def __init__(self, reason: str) -> None:
        super().__init__(reason)
        self.is_ok = False



from petch.betterlist import BetterList
from petch.betterdict import BetterDict
from petch.betterasync import (
    Producer,
    Consumer,
    MessageChannel,
    Message,
    run_concurrent_tasks,
)
from petch.decorators import perform_profiling, measure_time, debug, console
from petch.utils import compose, Ok, Err



from unittest import TestCase
from betterdict import BetterDict as BD


class TestBetterDict(TestCase):
    def test_keys_list(self):
        dict1 = BD({"a": 3, "b": 2, "c": 1})
        self.assertEqual(dict1.keys_list[0], "a")



import cProfile
import pstats
from time import time
from rich.console import Console
from rich.traceback import install
from rich.theme import Theme

console = Console(
    theme=Theme(
        {
            "info": "white",
            "warning": "yellow",
            "danger": "bold red",
            "debug": "green",
        }
    )
)
install()


def measure_time(func):
    def wrap_func(*args, **kwargs):
        t1 = time()
        result = func(*args, **kwargs)
        t2 = time()
        print(f"Function {func.__name__!r} executed in {(t2-t1):.4f}s")
        return result

    return wrap_func


def perform_profiling(func):
    def wrapper(*args, **kwargs):
        with cProfile.Profile() as pr:
            result = func(*args, **kwargs)
            stats = pstats.Stats(pr)
        stats.sort_stats(pstats.SortKey.TIME)
        stats.dump_stats(filename="profiling.prof")
        return result

    return wrapper


def debug(func):
    def wrap(*args, **kwargs):
        console.log(func.__name__, log_locals=True, style="debug")
        results = func(*args, **kwargs)
        return results

    return wrap



from typing import Any, Callable
from functools import reduce


def create_list(n: int):
    return BetterList(range(n))


class BetterList(list):
    """
    This List Class allows you to do javascript styled continuous operation.
    ex.
    x = BetterList([1,2,3,4])
    y = x.filter(func1).map(func2).reduce(func3)
    z = x.map(func4).for_each(func5)
    """

    @property
    def length(self):
        return len(self)

    def __add__(self, other):
        result = list(self) + list(other)
        return BetterList(result)

    def __mul__(self, n):
        result = list(self) * n
        return BetterList(result)

    def __getitem__(self, s):
        result = list(self)[s]
        if type(result) == list:
            return BetterList(result)
        return result

    def copy(self):
        return BetterList(super().copy())

    def sort(self, func: Callable = lambda x: x):
        return BetterList(sorted(self, key=func))

    def min(self, n_smallest: int = 1, func: Callable = lambda x: x):
        """
        Return looks like this:
        the min one is the first item
        [0, 0, 4, 5, 5, 5, 6, 10, 11, 12]
        """
        return BetterList(self.sort(func)[:n_smallest])

    def max(self, n_largest: int = 1, func: Callable = lambda x: x):
        """
        Return looks like this:
        the max one is the first item
        [100, 99, 98, 97, 96, 95, 94, 93, 93, 92]
        """
        return BetterList(self.sort(func)[-n_largest:][::-1])

    def filter(self, func: Callable):
        return BetterList(filter(func, self))

    def filter_out(self, func: Callable):
        return BetterList(filter(lambda x: not (func(x)), self))

    def remove_duplicate(self):
        return BetterList(set(self))

    def map(self, func: Callable):
        return BetterList(map(func, self))

    def reduce(self, func: Callable):
        return reduce(func, self)

    def foreach(self, func: Callable):
        for x in self:
            func(x)

    def filter_index(self, func: Callable):
        _index = BetterList([i for i, _ in enumerate(self)])
        return BetterList([self[i] for i in _index.filter(func)])

    def get(self, index, default_value=None):
        try:
            return self[index]
        except IndexError:
            return default_value

    def value_after(self, value: Any):
        it = iter(self)
        x = ""
        while str(x) != str(value) or str(x) not in str(value):
            x = next(it)
        return next(it)

    def first(self, n: int):
        return BetterList(range(0, n)).map(lambda i: self[i])

    def last(self, n: int):
        return BetterList(range(-n, 0)).map(lambda i: self[i])



from unittest import IsolatedAsyncioTestCase
from betterasync import MessageChannel, Producer, Consumer, run_concurrent_tasks


class TestBetterAsync(IsolatedAsyncioTestCase):
    async def test_system(self):
        class MyProducer(Producer):
            async def run(self):
                for i in range(3):
                    await self.send_msg("Normal", i)
                await self.send_err("End", "")
                return i

        class MyConsumer(Consumer):
            def __init__(self, name=""):
                super().__init__(name)
                self.x = 0

            async def on_msg(self, msg):
                if not msg.ok:
                    self.x += 10
                else:
                    self.x += 1

        channel = MessageChannel()
        producer_1 = MyProducer("Producer#1")
        producer_2 = MyProducer("Producer#2")
        consumer_1 = MyConsumer("Consumer#1")
        results = (
            channel.register_producer(producer_1)
            .register_producer(producer_2)
            .register_consumer(consumer_1)
            .run()
        )
        self.assertEqual(consumer_1.x, 26)
        self.assertEqual(results, [2, 2])

    def test_run_concurrent_tasks(self):
        async def my_task(a):
            return a**2

        results = run_concurrent_tasks(my_task, [1, 2, 3])
        self.assertEqual(results, [1, 4, 9])
